# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Set, List, Dict

from ..types import URI


class LiveTranslator(ABC):

    @abstractmethod
    def translate(self, text: str) -> Optional[str]:
        raise NotImplemented


class M3UInfo:

    def __init__(self, group: str, name: str, url: URI):
        super().__init__()
        self.__group = group
        self.__name = name
        self.__url = url

    @property
    def group(self) -> str:
        return self.__group

    @property
    def name(self) -> str:
        return self.__name

    @property
    def url(self) -> URI:
        return self.__url


class M3UTranslator(LiveTranslator):

    # noinspection PyMethodMayBeStatic
    def check(self, text: str) -> bool:
        return check_m3u(text=text)

    # noinspection PyMethodMayBeStatic
    def _split_items(self, text: str) -> List[str]:
        return split_m3u(text=text)

    # noinspection PyMethodMayBeStatic
    def _parse_info(self, lines: str) -> Optional[M3UInfo]:
        return parse_inf(lines=lines)

    def _parse_items(self, array: List[str]) -> Dict[str, Dict[str, Set[URI]]]:
        genres = {}
        for item in array:
            info = self._parse_info(item)
            if info is None:
                continue
            # get channels with group title
            channels: Dict = genres.get(info.group)
            if channels is None:
                channels = {}
                genres[info.group] = channels
            # get streams with channel name
            streams: Set = channels.get(info.name)
            if streams is None:
                streams = set()
                channels[info.name] = streams
            # add stream url
            streams.add(info.url)
        return genres

    # noinspection PyMethodMayBeStatic
    def _build_channel_line(self, name: str, streams: Set[URI]) -> str:
        text = '#'.join(streams)
        if len(text) < 2:
            return ''
        return '%s,%s' % (name, text)

    def _build_genre_block(self, title: str, channels: Dict[str, Set[URI]]) -> str:
        text = '\n'
        for name in channels:
            streams = channels[name]
            line = self._build_channel_line(name=name, streams=streams)
            text += '%s\n' % line
        if len(text) < 2:
            return ''
        return '%s,#genre#\n%s' % (title, text)

    # Override
    def translate(self, text: str) -> Optional[str]:
        if self.check(text=text):
            array = self._split_items(text=text)
            genres = self._parse_items(array=array)
            blocks: List[str] = []
            for title in genres:
                channels = genres[title]
                text = self._build_genre_block(title=title, channels=channels)
                blocks.append(text)
            if len(blocks) > 0:
                return '\n'.join(blocks)


"""
#EXTM3U x-tvg-url="https://live.x.com/e.xml"
#EXTINF:-1 tvg-name="CCTV1" tvg-logo="https://live.x.com/CCTV1.png" group-title="央视频道",CCTV-1 综合
http://1.2.3.4:81/CCTV1/index.m3u8
#EXTINF:-1 tvg-name="CCTV2" tvg-logo="https://live.x.com/CCTV2.png" group-title="央视频道",CCTV-2 财经
http://1.2.3.4:81/CCTV2/index.m3u8
...
"""


def check_m3u(text: str) -> bool:
    if text.find(r'#EXTM3U') < 0:
        return False
    elif text.find(r'#EXTINF') < 0:
        return False
    elif text.find(r' tvg-name="') < 0:
        return False
    elif text.find(r' group-title="') < 0:
        return False
    else:
        return True


def split_m3u(text: str) -> List[str]:
    array = []
    pos = text.find(r'#EXTINF')
    while pos > 0:
        text = text[pos:]
        pos = text.find(r'#EXTINF', 7)
        if pos < 0:
            array.append(text)
            break
        else:
            array.append(text[:pos])
    return array


def parse_inf(lines: str) -> Optional[M3UInfo]:
    array = lines.splitlines()
    if len(array) < 2:
        return None
    head = array[0]
    # get stream url
    url = array[1].strip()
    if url.find(r'://') < 0:
        return None
    # get channel name
    pos = head.rfind(r'",')
    if pos > 0:
        pos += 2
        name = head[pos:]
    else:
        name = fetch_field(text=head, tag_start=' tvg-name="', tag_end='"')
        if name is None:
            return None
    # get group title
    group = fetch_field(text=head, tag_start=' group-title="', tag_end='"')
    if group is None or len(name) == 0:
        return None
    return M3UInfo(group=group, name=name, url=url)


def fetch_field(text: str, tag_start: str, tag_end: str) -> Optional[str]:
    start = text.find(tag_start)
    if start < 0:
        return None
    start += len(tag_start)
    end = text.find(tag_end, start)
    if end > start:
        return text[start:end]



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    TV Box
    ~~~~~~

    Live Stream
"""


from .stream import LiveStream
from .channel import LiveChannel
from .genre import LiveGenre

from .factory import LiveChecker, LiveCreator
from .factory import LiveFactory

from .scanner import LiveStreamScanner

from .parser import LiveParser

from .translate import LiveTranslator
# from .translate import M3UInfo
from .translate import M3UTranslator


__all__ = [

    'LiveStream', 'LiveChannel', 'LiveGenre',

    'LiveChecker', 'LiveCreator',
    'LiveFactory',

    'LiveStreamScanner',

    'LiveParser',

    'LiveTranslator',
    # 'M3UInfo',
    'M3UTranslator',

]



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Log Util
    ~~~~~~~~
"""

import time


DEBUG_FLAG = 0x01
INFO_FLAG = 0x02
WARNING_FLAG = 0x04
ERROR_FLAG = 0x08


class Log:

    DEBUG = 0xFF    # 0000 1111 : debug(), info(), warning(), error()
    DEVELOP = 0xFE  # 0000 1110 :          info(), warning(), error()
    RELEASE = 0xFC  # 0000 1100 :                  warning(), error()

    LEVEL = RELEASE

    @classmethod
    def debug(cls, msg: str):
        if cls.LEVEL & DEBUG_FLAG == 0:
            return None
        print('[%s]  DEBUG  | %s' % (current_time(), msg))

    @classmethod
    def info(cls, msg: str):
        if cls.LEVEL & INFO_FLAG == 0:
            return None
        print('[%s]   LOG   | %s' % (current_time(), msg))

    @classmethod
    def warning(cls, msg: str):
        if cls.LEVEL & WARNING_FLAG == 0:
            return None
        print('[%s] WARNING | %s' % (current_time(), msg))

    @classmethod
    def error(cls, msg: str):
        if cls.LEVEL & ERROR_FLAG == 0:
            return None
        print('[%s]  ERROR  | %s' % (current_time(), msg))


class Logging:

    def debug(self, msg: str):
        Log.debug(msg='%s >\t%s' % (self.__class__.__name__, msg))

    def info(self, msg: str):
        Log.info(msg='%s >\t%s' % (self.__class__.__name__, msg))

    def warning(self, msg: str):
        Log.warning(msg='%s >\t%s' % (self.__class__.__name__, msg))

    def error(self, msg: str):
        Log.error(msg='%s >\t%s' % (self.__class__.__name__, msg))


def current_time() -> str:
    now = DateTime.current_timestamp()
    return DateTime.full_string(timestamp=now)


class DateTime:

    @classmethod
    def current_timestamp(cls) -> float:
        return time.time()

    @classmethod
    def full_string(cls, timestamp: float) -> str:
        localtime = time.localtime(timestamp)
        return time.strftime('%Y-%m-%d %H:%M:%S', localtime)



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Set, Tuple, List

from ..types import URI
from .stream import LiveStream
from .factory import LiveFactory
from .channel import LiveChannel
from .genre import LiveGenre


class LiveParser:

    def __init__(self):
        super().__init__()
        self.__factory = LiveFactory()

    @property
    def factory(self) -> LiveFactory:
        return self.__factory

    def parse(self, text: str) -> List[LiveGenre]:
        """ parse 'lives.txt' file content """
        lines = text.splitlines()
        return self.parse_lines(lines=lines)

    # protected
    def parse_lines(self, lines: List[str]) -> List[LiveGenre]:
        all_groups: List[LiveGenre] = []
        current = self.factory.new_genre(title='')
        for item in lines:
            text = item.strip()
            if len(text) == 0:
                continue
            elif text.startswith(r'#'):
                continue
            elif text.startswith(r'//'):
                continue
            # 1. check group name
            genre = self._fetch_genre(text=text)
            if genre is not None:
                # alternate current group
                if not current.empty:
                    all_groups.append(current)
                current = genre
                continue
            # 2. check channel name
            channel, text = self._fetch_channel(text=text)
            if channel is None:
                # empty line?
                continue
            else:
                # split streams
                sources = text.split(r'#')
            # 3. create streams
            streams: Set[LiveStream] = set()
            for src in sources:
                m3u8 = self._fetch_stream(text=src)
                if m3u8 is not None:
                    streams.add(m3u8)
            channel.add_streams(streams=streams)
            current.add_channel(channel=channel)
        # add last group
        if not current.empty:
            all_groups.append(current)
        return all_groups

    def _fetch_genre(self, text: str) -> Optional[LiveGenre]:
        title = split_genre(text=text)
        if title is None:
            return None
        return self.factory.new_genre(title=title)

    def _fetch_channel(self, text: str) -> Tuple[Optional[LiveChannel], str]:
        name, body = split_channel(text=text)
        if name is None:
            return None, text
        channel = self.factory.new_channel(name=name)
        return channel, body

    def _fetch_stream(self, text: str) -> Optional[LiveStream]:
        url, label = split_stream(text=text)
        if url is None:
            return None
        return self.factory.new_stream(url=url, label=label)


def split_genre(text: str) -> Optional[str]:
    pos = text.find(',#genre#')
    if pos > 0:
        return text[:pos].strip()


def split_channel(text: str) -> Tuple[Optional[str], str]:
    pos = text.find(',http')
    if pos < 0:
        # not a channel line
        return None, text
    # fetch channel name
    name = text[:pos].strip()
    pos += 1  # skip ','
    return name, text[pos:]


def split_stream(text: str) -> Tuple[Optional[URI], Optional[str]]:
    url = None
    label = None
    pair = text.split(r'$')
    if len(pair) == 1:
        url = text.strip()
        url = LiveStream.parse_url(url=url)
    else:
        # assert len(pair) == 2
        first = pair[0].strip()
        second = pair[1].strip()
        # check for url
        if first.find(r'://') > 0:
            url = first
            label = second
        elif second.find(r'://') > 0:
            url = second
            label = first
    # done
    return url, label



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import asyncio
from typing import Optional, Union, Set, List, Dict

from .types import URI
from .utils import Logging, DateTime
from .utils import http_get_text
from .utils import text_file_read, text_file_write, json_file_write

from .lives import LiveStream, LiveChannel, LiveGenre

from .config import LiveConfig
from .scanner import ScanContext, ScanEventHandler


class SourceLoader:

    def __init__(self):
        super().__init__()
        self.__resources: Dict[str, str] = {}

    def clear_caches(self):
        self.__resources.clear()

    async def load_text(self, src: Union[str, URI]) -> Optional[str]:
        # 1. check caches
        res = self.__resources.get(src)
        if res is None:
            # 2. cache not found, try to load
            res = await _load(src=src)
            if res is None:
                res = ''  # place holder
            # 3. cache the result
            self.__resources[src] = res
        # OK
        if len(res) > 0:
            return res


async def _load(src: Union[str, URI]) -> Optional[str]:
    if src.find(r'://') > 0:
        return await http_get_text(url=src, timeout=64)
    else:
        return await text_file_read(path=src)


class PreScanner(Logging):

    BATCH = 50

    async def scan_genres(self, context: ScanContext, genres: List[LiveGenre]):
        """ scan all streams in these genres batch by batch """
        all_streams: Set[LiveStream] = set()
        for group in genres:
            channels = group.channels
            for item in channels:
                streams = item.streams
                for src in streams:
                    all_streams.add(src)
        count = len(all_streams)
        index = 0
        sources: Set[LiveStream] = set()
        self.info(msg='pre-scanning %d stream(s) in %d genre(s) ...' % (count, len(genres)))
        for src in all_streams:
            index += 1
            self.info(msg='pre-scanning stream (%d/%d) %s' % (index, count, src.url))
            sources.add(src)
            if len(sources) >= self.BATCH:
                await self.scan_streams(context=context, streams=sources)
                sources.clear()
        if len(sources) > 0:
            await self.scan_streams(context=context, streams=sources)

    # protected
    async def scan_streams(self, context: ScanContext, streams: Set[LiveStream]):
        timeout = context.timeout
        now = DateTime.current_timestamp()
        tasks = [src.check(now=now, timeout=timeout) for src in streams]
        done, _ = await asyncio.wait(tasks)
        results = [item.result() for item in done if item.result()]
        self.info(msg='pre-scanned %d stream(s) => %d result(s)' % (len(tasks), len(results)))


class LiveHandler(ScanEventHandler, Logging):
    """ Live Storage """

    def __init__(self, config: LiveConfig):
        super().__init__()
        self.__config = config
        self.__scanner = PreScanner()

    @property
    def config(self) -> LiveConfig:
        return self.__config

    @property
    def batch_scanner(self) -> PreScanner:
        return self.__scanner

    @batch_scanner.setter
    def batch_scanner(self, scanner: PreScanner):
        self.__scanner = scanner

    #
    #   Storage
    #

    async def update_index(self, container: Dict) -> bool:
        """ update 'tvbox.json' """
        path = self.config.get_output_index_path()
        lives = container.get('lives')
        if path is None or len(lives) == 0:
            self.warning(msg='ignore live index: %s => %s' % (path, lives))
            return False
        self.info(msg='saving index file: %d live urls -> %s' % (len(lives), path))
        return await json_file_write(path=path, container=container)

    async def update_lives(self, genres: List[LiveGenre], url: URI) -> bool:
        """ update 'lives.txt' """
        path = self.config.get_output_lives_path(url=url)
        if path is None or len(genres) == 0:
            self.warning(msg='ignore live urls: %s => %s' % (url, genres))
            return False
        # build lives text
        text = ''
        for grp in genres:
            block = self._build_genre_block(genre=grp)
            if block is None:
                self.warning(msg='empty genre: %s' % grp)
                continue
            text += '\n%s' % block
        # OK
        self.info(msg='saving lives (%d genres) into file: %s -> %s' % (len(genres), url, path))
        if len(text) > 0:
            return await text_file_write(path=path, text=text)

    # format: '{title},#genre#\n{lines}'
    def _build_genre_block(self, genre: LiveGenre) -> Optional[str]:
        block = ''
        title = genre.title
        channels = genre.channels
        for item in channels:
            line = self._build_channel_line(channel=item)
            if line is None:
                self.warning(msg='empty channel: "%s", %s' % (title, item))
                continue
            block += '%s\n' % line
        # OK
        if len(block) > 0:
            block = '%s,#genre#\n%s' % (title, block)
            return block

    # format: '{name},{url}${tag}#{url}${tag}#...'
    def _build_channel_line(self, channel: LiveChannel) -> Optional[str]:
        line = ''
        name = channel.name
        streams = channel.streams
        for src in streams:
            fra = self._build_stream_fragment(stream=src)
            if fra is None:
                self.warning(msg='invalid stream: "%s", %s' % (name, src))
                continue
            line += '#%s' % fra
        # OK
        if len(line) > 1:
            line = line[1:]  # erase first '#'
            return '%s,%s' % (name, line)

    # format: '{url}${tag}'
    def _build_stream_fragment(self, stream: LiveStream) -> Optional[str]:
        if stream.available:
            url = stream.url
            label = stream.label
            if url is None or url.find(r'://') < 0:
                self.error(msg='stream error: %s' % stream)
            elif label is None or len(label) == 0:
                return url
            else:
                return '%s$%s' % (url, label)

    async def update_source(self, text: str, url: URI) -> bool:
        """ save source file for lives """
        path = self.config.get_output_source_path(url=url)
        if path is None:
            self.warning(msg='ignore index file: %s => %s' % (url, text))
            return False
        self.info(msg='saving source file: %d lines -> %s' % (len(text.splitlines()), path))
        if len(text) > 0:
            return await text_file_write(path=path, text=text)

    #
    #   ScanEventHandler
    #

    # Override
    async def on_scan_start(self, context: ScanContext, genres: List[LiveGenre]):
        # count all streams
        total_streams = 0
        for group in genres:
            channels = group.channels
            for item in channels:
                total_streams += len(item.streams)
        context.set(key='stream_total_count', value=total_streams)
        # pre scanning
        scanner = self.batch_scanner
        await scanner.scan_genres(context=context, genres=genres)

    # Override
    async def on_scan_finished(self, context: ScanContext, genres: List[LiveGenre]):
        self.info(msg='Mission accomplished.')

    # genre events

    # Override
    async def on_scan_genre_start(self, context: ScanContext, genre: LiveGenre):
        pass

    # Override
    async def on_scan_genre_finished(self, context: ScanContext, genre: LiveGenre):
        pass

    # channel events

    # Override
    async def on_scan_channel_start(self, context: ScanContext, genre: LiveGenre, channel: LiveChannel):
        pass

    # Override
    async def on_scan_channel_finished(self, context: ScanContext, genre: LiveGenre, channel: LiveChannel):
        pass

    # stream events

    # Override
    async def on_scan_stream_start(self, context: ScanContext, channel: LiveChannel, stream: LiveStream):
        pass

    # Override
    async def on_scan_stream_finished(self, context: ScanContext, channel: LiveChannel, stream: LiveStream):
        offset = context.get(key='stream_offset', default=0)
        total = context.get(key='stream_total_count', default=0)
        self.info(msg='Scanned (%d/%d) stream: "%s"\t-> %s' % (offset + 1, total, channel.name, stream))



# -*- coding: utf-8 -*-
#
#   TV-Box
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    TV Box Scanner
    ~~~~~~~~~~~~~~

"""

import getopt
import os
import sys

path = os.path.abspath(__file__)
path = os.path.dirname(path)
path = os.path.dirname(path)
sys.path.insert(0, path)

from tvbox.utils import Log
from tvbox.utils import AsyncRunner as Runner
from tvbox.lives import M3UTranslator
from tvbox.config import LiveConfig
from tvbox.source import LiveHandler
from tvbox.loader import LiveLoader
from tvbox.scanner import LiveScanner, ScanContext, ScanParser


#
# show logs
#
Log.LEVEL = Log.DEVELOP


DEFAULT_CONFIG = '/etc/tvbox/config.json'


def show_help():
    cmd = sys.argv[0]
    print('')
    print('    TV Box Scanner')
    print('')
    print('usages:')
    print('    %s [--config=<FILE>] scan lives' % cmd)
    print('    %s [-h|--help]' % cmd)
    print('')
    print('actions:')
    print('    scan lives      scan for live stream sources')
    print('')
    print('optional arguments:')
    print('    --config        config file path (default: "%s")' % DEFAULT_CONFIG)
    print('    --help, -h      show this help message and exit')
    print('')


async def async_main():
    try:
        opts, args = getopt.getopt(args=sys.argv[1:],
                                   shortopts='hf:',
                                   longopts=['help', 'config='])
    except getopt.GetoptError:
        show_help()
        sys.exit(1)
    # check options
    config_file = None
    for opt, arg in opts:
        if opt == '--config':
            config_file = arg
        else:
            show_help()
            sys.exit(0)
    # check config file path
    if config_file is None:
        config_file = DEFAULT_CONFIG
    if not os.path.exists(config_file):
        show_help()
        Log.error(msg='config file not exists: %s' % config_file)
        sys.exit(0)
    # load config
    config = await LiveConfig.load(path=config_file)
    # initializing
    Log.info(msg='!!!')
    Log.info(msg='!!! Init with config: %s => %s' % (config_file, config))
    Log.info(msg='!!!')
    # create scanner
    translator = M3UTranslator()
    parser = ScanParser(translators=[translator])
    scanner = LiveScanner(parser=parser)
    # create loader
    loader = LiveLoader(config=config, scanner=scanner)
    await loader.load(handler=LiveHandler(config=config),
                      context=ScanContext(timeout=64))
    Log.info(msg='!!!')
    Log.info(msg='!!! Mission Accomplished.')
    Log.info(msg='!!!')


def main():
    Runner.sync_run(main=async_main())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, List, Dict

from .types import MapInfo
from .utils import Logging

from .lives import LiveStream, LiveChannel, LiveGenre
from .lives import LiveStreamScanner
from .lives import LiveParser, LiveTranslator


class ScanContext(MapInfo):

    def __init__(self, info: Dict = None, timeout: float = None):
        super().__init__(info=info)
        if timeout is not None and timeout > 0:
            self.set(key='timeout', value=timeout)

    @property
    def timeout(self) -> Optional[float]:
        return self.get(key='timeout')

    @timeout.setter
    def timeout(self, duration: float):
        self.set(key='timeout', value=duration)

    @property
    def cancelled(self) -> bool:
        flag = self.get(key='cancelled')
        if flag is None:
            flag = self.get(key='stopped', default=False)
        return flag

    @cancelled.setter
    def cancelled(self, flag: bool):
        self.set(key='cancelled', value=flag)


class ScanEventHandler(ABC):
    """ Scan callback """

    @abstractmethod
    async def on_scan_start(self, context: ScanContext, genres: List[LiveGenre]):
        raise NotImplemented

    @abstractmethod
    async def on_scan_finished(self, context: ScanContext, genres: List[LiveGenre]):
        raise NotImplemented

    # genre events

    @abstractmethod
    async def on_scan_genre_start(self, context: ScanContext, genre: LiveGenre):
        raise NotImplemented

    @abstractmethod
    async def on_scan_genre_finished(self, context: ScanContext, genre: LiveGenre):
        raise NotImplemented

    # channel events

    @abstractmethod
    async def on_scan_channel_start(self, context: ScanContext, genre: LiveGenre, channel: LiveChannel):
        raise NotImplemented

    @abstractmethod
    async def on_scan_channel_finished(self, context: ScanContext, genre: LiveGenre, channel: LiveChannel):
        raise NotImplemented

    # stream events

    @abstractmethod
    async def on_scan_stream_start(self, context: ScanContext, channel: LiveChannel, stream: LiveStream):
        raise NotImplemented

    @abstractmethod
    async def on_scan_stream_finished(self, context: ScanContext, channel: LiveChannel, stream: LiveStream):
        raise NotImplemented


class ScanParser(LiveParser, Logging):

    def __init__(self, translators: List[LiveTranslator]):
        super().__init__()
        self.__translators = translators

    @property
    def translators(self) -> List[LiveTranslator]:
        return self.__translators

    def _translate(self, text: str) -> Optional[str]:
        translators = self.__translators
        for trans in translators:
            target = trans.translate(text=text)
            if target is None:
                continue
            # elif len(target) < 8:
            #     continue
            return target

    # Override
    def parse(self, text: str) -> List[LiveGenre]:
        target = self._translate(text=text)
        if target is None:
            return super().parse(text=text)
        cnt1 = len(text.splitlines())
        cnt2 = len(target.splitlines())
        self.warning(msg='converted text from %d lines to %d lines' % (cnt1, cnt2))
        return super().parse(text=target)


class LiveScanner:

    def __init__(self, parser: LiveParser):
        super().__init__()
        self.__parser = parser
        self.__scanner = self._create_stream_scanner()

    # noinspection PyMethodMayBeStatic
    def _create_stream_scanner(self) -> LiveStreamScanner:
        # TODO: override for customized scanner
        return LiveStreamScanner()

    @property  # protected
    def parser(self) -> LiveParser:
        return self.__parser

    @property  # protected
    def stream_scanner(self) -> LiveStreamScanner:
        return self.__scanner

    async def scan(self, text: str, context: ScanContext, handler: ScanEventHandler) -> List[LiveGenre]:
        """ Get non-empty channel groups """
        all_genres = self.parser.parse(text=text)
        # reset pointers
        context.set(key='genre_offset', value=0)
        context.set(key='channel_offset', value=0)
        context.set(key='stream_offset', value=0)
        # start mission
        await handler.on_scan_start(context=context, genres=all_genres)
        groups = await self._scan_genres(genres=all_genres, context=context, handler=handler)
        # mission accomplished
        await handler.on_scan_finished(context=context, genres=all_genres)
        return groups

    async def _scan_genres(self, genres: List[LiveGenre],
                           context: ScanContext, handler: ScanEventHandler) -> List[LiveGenre]:
        """ Get non-empty genres """
        groups: List[LiveGenre] = []
        context.set(key='genre_count', value=len(genres))
        # positions
        offset = context.get(key='genre_offset', default=0)
        index = 0
        for item in genres:
            if context.cancelled:
                break
            else:
                context.set(key='genre_index', value=index)
            # scan genre start
            await handler.on_scan_genre_start(context=context, genre=item)
            channels = await self._scan_channels(channels=item.channels, genre=item, context=context, handler=handler)
            item.channels = channels
            if len(channels) > 0:
                groups.append(item)
            # scan genre finished
            await handler.on_scan_genre_finished(context=context, genre=item)
            index += 1
            offset += 1
            context.set(key='genre_offset', value=offset)
        return groups

    async def _scan_channels(self, channels: List[LiveChannel], genre: LiveGenre,
                             context: ScanContext, handler: ScanEventHandler) -> List[LiveChannel]:
        """ Get available channels in this genre """
        available_channels: List[LiveChannel] = []
        context.set(key='channel_count', value=len(channels))
        # positions
        offset = context.get(key='channel_offset', default=0)
        index = 0
        for item in channels:
            if context.cancelled:
                break
            else:
                context.set(key='channel_index', value=index)
            # scan channel start
            await handler.on_scan_channel_start(context=context, genre=genre, channel=item)
            streams = await self._scan_streams(streams=item.streams, channel=item, context=context, handler=handler)
            item.streams = streams
            if len(streams) > 0:
                available_channels.append(item)
            # scan channel finished
            await handler.on_scan_channel_finished(context=context, genre=genre, channel=item)
            index += 1
            offset += 1
            context.set(key='channel_offset', value=offset)
        return available_channels

    async def _scan_streams(self, streams: List[LiveStream], channel: LiveChannel,
                            context: ScanContext, handler: ScanEventHandler) -> List[LiveStream]:
        """ Get available streams in this channel """
        available_streams: List[LiveStream] = []
        context.set(key='stream_count', value=len(streams))
        # positions
        offset = context.get(key='stream_offset', default=0)
        index = 0
        for item in streams:
            if context.cancelled:
                break
            else:
                context.set(key='stream_index', value=index)
            # scan stream start
            await handler.on_scan_stream_start(context=context, channel=channel, stream=item)
            src = await self.stream_scanner.scan_stream(stream=item, timeout=context.timeout)
            if src is not None:  # and src.available:
                available_streams.append(src)
            # scan stream finished
            await handler.on_scan_stream_finished(context=context, channel=channel, stream=item)
            index += 1
            offset += 1
            context.set(key='stream_offset', value=offset)
        return available_streams



# -*- coding: utf-8 -*-
#
#   TV-Box
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    TV Box
    ~~~~~~

"""


from .scanner import ScanContext, ScanEventHandler
from .scanner import ScanParser
from .scanner import LiveScanner

from .config import LiveConfig
from .source import SourceLoader, LiveHandler
from .loader import LiveLoader


name = 'TV-Box'

__author__ = 'Albert Moky'


__all__ = [

    'ScanContext', 'ScanEventHandler',
    'ScanParser',
    'LiveScanner',

    'LiveConfig',
    'SourceLoader', 'LiveHandler',
    'LiveLoader',

]



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import asyncio
import threading
from abc import ABC, abstractmethod
from typing import Coroutine

from .log import Logging


class AsyncRunner(Logging, ABC):
    """ Daemon Runner """

    def __init__(self, interval: float):
        super().__init__()
        assert interval > 0, 'interval error: %s' % interval
        self.__interval = interval
        self.__running = False

    @property
    def interval(self) -> float:
        return self.__interval

    @property
    def running(self) -> bool:
        return self.__running

    async def start(self):
        self.__running = True
        await self.run()

    async def stop(self):
        self.__running = False

    # protected
    async def run(self):
        while self.running:
            try:
                ok = await self.process()
            except Exception as error:
                self.error(msg='process failed: %s' % error)
                ok = False
            if ok:
                # runner is busy, return True to go on.
                continue
            else:
                # if nothing to do now, return False here
                # to let the thread have a rest.
                await self._idle()

    @abstractmethod  # protected
    async def process(self) -> bool:
        raise NotImplemented

    # protected
    async def _idle(self):
        await self.sleep(seconds=self.interval)
        # time.sleep(self.interval)

    @classmethod
    async def sleep(cls, seconds: float):
        await asyncio.sleep(seconds)

    @classmethod
    def sync_run(cls, main: Coroutine):
        """ Run main coroutine until complete """
        return asyncio.run(main)

    @classmethod
    def async_task(cls, coro: Coroutine) -> asyncio.Task:
        """ Create an async task to run the coroutine """
        return asyncio.create_task(coro)

    @classmethod
    def async_thread(cls, coro: Coroutine) -> threading.Thread:
        """ Create a daemon thread to run the coroutine """
        return threading.Thread(target=cls.sync_run, args=(coro,), daemon=True)



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Union, Any, Set, List, Dict

from .types import URI
from .utils import Logging, DateTime
from .utils import json_decode
from .utils import path_parent, path_join

from .config import LiveConfig
from .scanner import LiveScanner, ScanContext
from .source import SourceLoader, LiveHandler


class LiveLoader(Logging):

    def __init__(self, config: LiveConfig, scanner: LiveScanner):
        super().__init__()
        self.__config = config
        self.__scanner = scanner
        self.__loader = self._create_source_loader()

    @property
    def config(self) -> LiveConfig:
        return self.__config

    @property
    def scanner(self) -> LiveScanner:
        return self.__scanner

    @property
    def loader(self) -> SourceLoader:
        return self.__loader

    # noinspection PyMethodMayBeStatic
    def _create_source_loader(self) -> SourceLoader:
        # TODO: override for customized loader
        return SourceLoader()

    def clear_caches(self):
        self.loader.clear_caches()

    #
    #   Load
    #

    async def get_live_urls(self) -> Set[URI]:
        live_urls = set()
        # 1. get from "sources"
        sources = self.config.sources
        self.info(msg='getting live urls from: %s' % sources)
        for idx in sources:
            src = _get_source_url(item=idx)
            if len(src) == 0:
                self.error(msg='source error: %s' % idx)
                continue
            # elif src.find(r'://') < 0:
            #     self.error(msg='source error: %s -> %s' % (idx, src))
            #     continue
            # 1.1. load
            text = await self._load_resource(src=src)
            if text is None:
                self.error(msg='ignore index: %s' % src)
                continue
            # 1.2. parse
            info = json_decode(string=text)
            if not isinstance(info, Dict):
                self.error(msg='json error: %s -> %s' % (src, text))
                continue
            # 1.3. get 'lives' from this source
            lives = info.get('lives')
            if not isinstance(lives, List):
                self.error(msg='json error: %s -> %s' % (src, text))
                continue
            # 1.4. add lives url
            for item in lives:
                url = _get_live_url(item=item)
                if len(url) == 0:
                    self.error(msg='lives item error: %s' % item)
                    continue
                elif url.find(r'://') < 0:
                    base = path_parent(src)
                    url = path_join(base, url)
                live_urls.add(url)
        # 2. get from "lives"
        lives = self.config.lives
        for item in lives:
            url = _get_live_url(item=item)
            if url.find(r'://') < 0:
                self.error(msg='lives item error: %s' % item)
                continue
            live_urls.add(url)
        # OK
        return live_urls

    async def _load_resource(self, src: Union[str, URI]) -> Optional[str]:
        self.info(msg='loading resource from: "%s"' % src)
        res = await self.loader.load_text(src=src)
        if res is None:
            self.error(msg='failed to load resource: "%s"' % src)
        return res

    #
    #   Running
    #

    async def load(self, handler: LiveHandler, context: ScanContext):
        scanner = self.scanner
        available_lives: List[Dict] = []
        # scan lives
        start_time = DateTime.current_timestamp()
        live_urls = await self.get_live_urls()
        self.info(msg='got live urls: %d, %s' % (len(live_urls), live_urls))
        for url in live_urls:
            text = await self._load_resource(src=url)
            if text is None:
                self.error(msg='ignore lives: %s' % url)
                continue
            else:
                await handler.update_source(text=text, url=url)
            count = len(text.splitlines())
            self.info(msg='scanning lives: %s => %d lines' % (url, count))
            genres = await scanner.scan(text=text, context=context, handler=handler)
            if len(genres) == 0:
                self.warning(msg='ignore empty lives: %s => %d lines' % (url, count))
                continue
            else:
                # update 'lives.txt'
                await handler.update_lives(genres=genres, url=url)
            # append to 'lives'
            available_lives.append({
                'url': self.config.get_output_lives_url(url=url),
                'path': self.config.get_output_lives_path(url=url),
                'src': url,
            })
        # mission accomplished
        end_time = DateTime.current_timestamp()
        return await handler.update_index(container={
            'lives': available_lives,
            'scan_time': DateTime.full_string(timestamp=start_time),
            'update_time': DateTime.full_string(timestamp=end_time),
        })


def _get_source_url(item: Any) -> str:
    if isinstance(item, Dict):
        item = item.get('url')
    if isinstance(item, str):
        return item
    else:
        return ''


def _get_live_url(item: Any) -> str:
    if isinstance(item, Dict):
        item = item.get('url')
    if isinstance(item, str):
        return item
    else:
        return ''



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

from dimples.utils import Log, Runner
from dimples.utils import Path

path = Path.abs(path=__file__)
path = Path.dir(path=path)
path = Path.dir(path=path)
Path.add(path=path)

from tvbox.utils import http_get_text
from tvbox.lives import M3UTranslator
from tvbox import ScanParser


#
# show logs
#
Log.LEVEL = Log.DEVELOP


async def async_main():
    url = "https://live.fanmingming.com/tv/m3u/ipv6.m3u"
    text = await http_get_text(url=url)
    if text is None:
        Log.error(msg='failed to download url: %s' % url)
        return False
    else:
        count = len(text.splitlines())
        Log.info(msg='download url: %s -> %d lines' % (url, count))
    translator = M3UTranslator()
    # lives_txt = translator.translate(text=text)
    # Log.info(msg='translated: \n%s' % lives_txt)
    parser = ScanParser(translators=[translator])
    genres = parser.parse(text=text)
    Log.info(msg='parsed: \n%s' % genres)


def main():
    Runner.sync_run(main=async_main())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import threading
from typing import Optional, Any, List, Dict
from typing import Iterable

from ..types import URI, MapInfo
from ..utils import DateTime


class LiveStream(MapInfo):
    """ M3U8 """

    SCAN_INTERVAL = 3600 * 2

    def __init__(self, info: Dict = None, url: URI = None, label: str = None):
        """
        Create Live Stream

        :param info: stream info
        :param url:  m3u8
        """
        super().__init__(info=info)
        if url is not None:
            self.set(key='url', value=url)
        if label is not None:
            self.set(key='label', value=label)
        # checker lock
        self.__lock = threading.Lock()

    # Override
    def __str__(self) -> str:
        cname = self.__class__.__name__
        when = self.time
        if when is not None:
            when = DateTime.full_string(timestamp=when)
        return '<%s time="%s" ttl=%s label="%s" url="%s" />' % (cname, when, self.ttl, self.label, self.url)

    # Override
    def __repr__(self) -> str:
        cname = self.__class__.__name__
        when = self.time
        if when is not None:
            when = DateTime.full_string(timestamp=when)
        return '<%s time="%s" ttl=%s label="%s" url="%s" />' % (cname, when, self.ttl, self.label, self.url)

    # Override
    def __hash__(self) -> int:
        """ Return hash(self). """
        return self.url.__hash__()

    # Override
    def __eq__(self, x: str) -> bool:
        """ Return self==value. """
        if isinstance(x, LiveStream):
            if self is x:
                # same object
                return True
            x = x.url
        # check url
        return self.url.__eq__(x)

    # Override
    def __ne__(self, x: str) -> bool:
        """ Return self!=value. """
        if isinstance(x, LiveStream):
            if self is x:
                # same object
                return False
            x = x.url
        # check url
        return self.url.__ne__(x)

    @property
    def url(self) -> URI:
        """ m3u8 """
        return self.get(key='url', default='')

    @property
    def label(self) -> Optional[str]:
        """ name """
        return self.get(key='label', default=None)

    @property
    def empty(self) -> bool:
        url = self.get(key='url', default=None)
        return url is None or len(url) == 0

    @property
    def available(self) -> bool:
        ttl = self.get(key='ttl', default=0)
        return ttl > 0

    @property
    def ttl(self) -> Optional[float]:
        """ time to load (network speed) """
        return self.get(key='ttl', default=None)

    @property
    def time(self) -> Optional[float]:
        return self.get(key='time', default=None)

    def set_ttl(self, ttl: float, now: float = None):
        """ update speed """
        if now is None:
            now = DateTime.current_timestamp()
        self.set(key='ttl', value=ttl)
        self.set(key='time', value=now)

    def is_expired(self, now: float = None) -> bool:
        """ whether needs to check """
        last_time = self.time
        if last_time is None:
            # never checked yet
            return True
        elif now is None:
            now = DateTime.current_timestamp()
        # check interval
        return now > (last_time + self.SCAN_INTERVAL)

    async def check(self, now: float = None, timeout: float = None) -> bool:
        """ check whether ttl > 0 """
        with self.__lock:
            if self.is_expired(now=now):
                # check it again
                ttl = await _stream_checker().check_stream(stream=self, timeout=timeout)
            else:
                # no need to check again now
                ttl = self.ttl
            return ttl is not None and ttl > 0

    #
    #   Factories
    #

    @classmethod
    def parse(cls, info: Any):  # -> Optional[LiveStream]:
        if info is None:
            return None
        elif isinstance(info, LiveStream):
            return info
        elif isinstance(info, MapInfo):
            info = info.dictionary
        # if 'url' in info:
        #     return cls(info=info)
        return _stream_factory().create_stream(info=info)

    @classmethod
    def convert(cls, array: Iterable[Dict]):  # -> List[LiveStream]:
        streams: List[LiveStream] = []
        for item in array:
            info = cls.parse(info=item)
            if info is not None:
                streams.append(info)
        return streams

    @classmethod
    def revert(cls, streams: Iterable) -> List[Dict]:
        array: List[Dict] = []
        for item in streams:
            if isinstance(item, MapInfo):
                array.append(item.dictionary)
            elif isinstance(item, Dict):
                array.append(item)
        return array

    @classmethod
    def parse_url(cls, url: str) -> Optional[URI]:
        if not isinstance(url, str):
            return None
        if url.find(r'://') > 0:
            return url


def _stream_factory():
    from .factory import LiveFactory
    return LiveFactory()


def _stream_checker():
    factory = _stream_factory()
    return factory.checker



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from .crypto import md5
from .crypto import utf8_encode, utf8_decode
from .crypto import hex_encode, hex_decode
from .crypto import json_encode, json_decode

from .aio import http_head, http_get
from .aio import http_get_text, http_check_m3u8

from .aio import path_parent, path_join
from .aio import text_file_read, text_file_write
from .aio import json_file_read, json_file_write

from .log import DateTime
from .log import Log, Logging

from .runner import AsyncRunner


class Singleton(object):

    __instances = {}

    def __init__(self, cls):
        self.__cls = cls

    def __call__(self, *args, **kwargs):
        cls = self.__cls
        instance = self.__instances.get(cls, None)
        if instance is None:
            instance = cls(*args, **kwargs)
            self.__instances[cls] = instance
        return instance

    def __getattr__(self, key):
        return getattr(self.__cls, key, None)


__all__ = [

    'md5',
    'utf8_encode', 'utf8_decode',
    'hex_encode', 'hex_decode',
    'json_encode', 'json_decode',

    'http_head', 'http_get',
    'http_get_text', 'http_check_m3u8',

    'path_parent', 'path_join',
    'text_file_read', 'text_file_write',
    'json_file_read', 'json_file_write',

    'DateTime',
    'Log', 'Logging',

    'AsyncRunner',

    'Singleton',

]



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import threading
from typing import Optional, List

from ..utils import Logging
from ..utils import DateTime
from ..utils import AsyncRunner as Runner
from .stream import LiveStream


class LiveStreamScanner(Runner, Logging):

    def __init__(self):
        super().__init__(interval=2)
        # caches
        self.__queue: List[LiveStream] = []
        self.__lock = threading.Lock()
        # background thread
        thr = Runner.async_thread(coro=self.start())
        thr.start()

    async def scan_stream(self, stream: LiveStream, timeout: float = None) -> Optional[LiveStream]:
        """
        Check whether the stream source is available

            1. when it's available and not expired,
               return itself immediately;
            2. when it's checked before, add an update task in background thread,
               and return the stream itself immediately;
            3. otherwise, check the url, if respond nothing/error,
               return None, else return the stream.

        :param stream: live stream
        :param timeout:
        :return: None on invalid
        """
        now = DateTime.current_timestamp()
        with self.__lock:
            if not stream.is_expired(now=now):
                # this stream was checked recently,
                # no need to check again.
                is_first = False
            elif stream.time is not None:
                # this stream was checked before, but expired now; so
                # add it in the waiting queue to check it in background thread.
                self.__queue.append(stream)
                is_first = False
            else:
                # first check
                is_first = True
        if is_first:
            await stream.check(now=now, timeout=timeout)
        if stream.available:
            return stream

    # Override
    async def process(self) -> bool:
        with self.__lock:
            if len(self.__queue) == 0:
                # nothing to do now
                return False
            stream = self.__queue.pop(0)
        try:
            await stream.check(timeout=64)
        except Exception as error:
            self.error(msg='failed to scan stream: %s, error: %s' % (stream, error))



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Iterator, Any, Dict
from typing import ItemsView, KeysView, ValuesView


class MapInfo:

    def __init__(self, info: Dict = None):
        super().__init__()
        if info is None:
            info = {}
        self.__dictionary = info

    @property
    def dictionary(self) -> Dict:
        return self.__dictionary

    def clear(self):
        """ D.clear() -> None.  Remove all items from D. """
        self.__dictionary.clear()

    def get(self, key: str, default: Any = None) -> Any:
        """ Return the value for key if key is in the dictionary, else default. """
        return self.__dictionary.get(key, default)

    def set(self, key: str, value: Optional[Any]):
        if value is None:
            self.__dictionary.pop(key, None)
        else:
            if isinstance(value, MapInfo):
                value = value.dictionary
            self.__dictionary[key] = value

    def items(self) -> ItemsView[str, Any]:
        """ D.items() -> a set-like object providing a view on D's items """
        return self.__dictionary.items()

    def keys(self) -> KeysView[str]:
        """ D.keys() -> a set-like object providing a view on D's keys """
        return self.__dictionary.keys()

    def values(self) -> ValuesView[Any]:
        """ D.values() -> an object providing a view on D's values """
        return self.__dictionary.values()

    def pop(self, key: str, default: Optional[Any] = None) -> Optional[Any]:
        """
        D.pop(k[,d]) -> v, remove specified key and return the corresponding value.
        If key is not found, d is returned if given, otherwise KeyError is raised
        """
        return self.__dictionary.pop(key, default)

    def __contains__(self, o) -> bool:
        """ True if the dictionary has the specified key, else False. """
        return self.__dictionary.__contains__(o)

    def __getitem__(self, k: str) -> Any:
        """ x.__getitem__(y) <==> x[y] """
        return self.__dictionary.__getitem__(k)

    def __setitem__(self, k: str, v: Optional[Any]):
        """ Set self[key] to value. """
        self.__dictionary.__setitem__(k, v)

    def __delitem__(self, v: str):
        """ Delete self[key]. """
        self.__dictionary.__delitem__(v)

    def __iter__(self) -> Iterator[str]:
        """ Implement iter(self). """
        return self.__dictionary.__iter__()

    def __repr__(self) -> str:
        """ Return repr(self). """
        return self.__dictionary.__repr__()

    def __sizeof__(self) -> int:
        """ D.__sizeof__() -> size of D in memory, in bytes """
        return self.__dictionary.__sizeof__()

    def __len__(self) -> int:
        """ Return len(self). """
        return self.__dictionary.__len__()

    def __eq__(self, o: Dict) -> bool:
        """ Return self==value. """
        if isinstance(o, MapInfo):
            if self is o:
                # same object
                return True
            o = o.dictionary
        # check inner map
        return self.__dictionary.__eq__(o)

    def __ne__(self, o: Dict) -> bool:
        """ Return self!=value. """
        if isinstance(o, MapInfo):
            if self is o:
                # same object
                return False
            o = o.dictionary
        # check inner map
        return self.__dictionary.__ne__(o)



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import threading
from typing import Optional, Dict

from ..types import URI
from ..utils import DateTime
from ..utils import Singleton
from ..utils import http_check_m3u8

from .stream import LiveStream
from .channel import LiveChannel
from .genre import LiveGenre


class LiveChecker:

    def clear_caches(self):
        pass

    # noinspection PyMethodMayBeStatic
    async def check_stream(self, stream: LiveStream, timeout: float = None) -> Optional[float]:
        return await _check_stream(stream=stream, timeout=timeout)


class LiveCreator:
    """ Customizable Creator """

    # noinspection PyMethodMayBeStatic
    def create_stream(self, info: Dict = None, url: URI = None, label: str = None) -> LiveStream:
        return LiveStream(info=info, url=url, label=label)

    # noinspection PyMethodMayBeStatic
    def create_channel(self, info: Dict = None, name: str = None) -> LiveChannel:
        return LiveChannel(info=info, name=name)

    # noinspection PyMethodMayBeStatic
    def create_genre(self, info: Dict = None, title: str = None) -> LiveGenre:
        return LiveGenre(info=info, title=title)


@Singleton
class LiveFactory:

    def __init__(self):
        super().__init__()
        self.__checker = LiveChecker()
        self.__creator = LiveCreator()
        # caches
        self.__streams: Dict[URI, LiveStream] = {}
        self.__lock = threading.Lock()

    def clear_caches(self):
        with self.__lock:
            self.__checker.clear_caches()
            self.__streams.clear()

    @property
    def checker(self) -> LiveChecker:
        return self.__checker

    @checker.setter
    def checker(self, delegate: LiveChecker):
        self.__checker = delegate

    @property
    def creator(self) -> LiveCreator:
        return self.__creator

    @creator.setter
    def creator(self, creator: LiveCreator):
        self.__creator = creator

    #
    #   Genre
    #

    def create_genre(self, info: Dict) -> Optional[LiveGenre]:
        title = info.get('title')
        if title is None:  # or len(title) == 0:
            return None
        return self.creator.create_genre(info=info)

    def new_genre(self, title: str) -> LiveGenre:
        # assert len(title) > 0, 'genre title should not be empty'
        return self.creator.create_genre(title=title)

    #
    #   Live Channel
    #

    def create_channel(self, info: Dict) -> Optional[LiveChannel]:
        name = info.get('name')
        if name is None or len(name) == 0:
            return None
        return self.creator.create_channel(info=info)

    def new_channel(self, name: str) -> LiveChannel:
        assert len(name) > 0, 'channel name should not be empty'
        return self.creator.create_channel(name=name)

    #
    #   Live Stream Source
    #

    def create_stream(self, info: Dict) -> Optional[LiveStream]:
        url = info.get('url')
        url = LiveStream.parse_url(url=url)
        if url is None:
            return None
        with self.__lock:
            src = self.__streams.get(url)
            if src is None:
                src = self.creator.create_stream(info=info)
                self.__streams[url] = src
            return src

    def new_stream(self, url: URI, label: Optional[str]) -> LiveStream:
        with self.__lock:
            src = self.__streams.get(url)
            if src is None:
                src = self.creator.create_stream(url=url, label=label)
                self.__streams[url] = src
            return src

    def set_stream(self, stream: LiveStream):
        url = stream.url
        with self.__lock:
            old = self.__streams.get(url)
            if old is not None:
                _merge_stream(new=stream, old=old)
            self.__streams[url] = stream


def _merge_stream(new: LiveStream, old: LiveStream) -> LiveStream:
    if new is old:
        return new
    new_time = new.time
    old_time = old.time
    if new_time is None and old_time is None:
        # neither of them were checked
        return new
    elif new_time is None:
        # copy old info
        new.set_ttl(ttl=old.ttl, now=old_time)
    elif old_time is None:
        # copy new info
        old.set_ttl(ttl=new.ttl, now=new_time)
    elif new_time < old_time:
        # copy old info
        new.set_ttl(ttl=old.ttl, now=old_time)
    elif old_time < new_time:
        # copy new info
        old.set_ttl(ttl=new.ttl, now=new_time)
    # OK
    return new


async def _check_stream(stream: LiveStream, timeout: float = None) -> Optional[float]:
    start_time = DateTime.current_timestamp()
    available = await http_check_m3u8(url=stream.url, timeout=timeout)
    end_time = DateTime.current_timestamp()
    if available:
        ttl = end_time - start_time
    else:
        ttl = None
    stream.set_ttl(ttl=ttl, now=end_time)
    return ttl



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
    TV Box
    ~~~~~~

    Live Stream
"""

import io

from setuptools import setup, find_packages

__version__ = '0.3.0'
__author__ = 'Albert Moky'
__contact__ = 'albert.moky@gmail.com'

with io.open('README.md', 'r', encoding='utf-8') as fh:
    readme = fh.read()

setup(
    name='tvbox',
    version=__version__,
    url='https://github.com/dimpart/chatbot-py',
    license='MIT',
    author=__author__,
    author_email=__contact__,
    description='TV Box: Live Stream',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    package_data={
        '': ['res/*.js']
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    entry_points={
        'console_scripts': [
            'tvbox=tvbox.run:main',
        ]
    },
    install_requires=[
        'aiohttp',     # 3.9.5
        # 'aiosignal',   # 1.3.1
        # 'attrs',       # 23.2.0
        # 'frozenlist',  # 1.4.1
        # 'multidict',   # 6.0.5
        # 'yarl',        # 1.9.4

        'aiou>=0.1.0',
    ]
)



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Union, Tuple, List, Dict, Mapping

import aiohttp
from aiou import Path, TextFile

from ..types import URI
from .crypto import json_encode, json_decode
from .log import Log


#
#   HTTP
#

async def http_head(url: URI, allow_redirects: bool = False, timeout: float = None) -> Tuple[int, Mapping]:
    """ Get HTTP status and headers """
    try:
        async with aiohttp.ClientSession() as session:
            async with session.head(url=url, allow_redirects=allow_redirects, timeout=timeout) as response:
                return response.status, response.headers
    except Exception as error:
        Log.error(msg='failed to head URL: %s, error: %s' % (url, error))
        return 404, {}


async def http_get(url: URI, allow_redirects: bool = False, timeout: float = None) -> Optional[bytes]:
    """ Get data from URL """
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url=url, allow_redirects=allow_redirects, timeout=timeout) as response:
                return await response.read()
    except Exception as error:
        Log.error(msg='failed to get URL: %s, error: %s' % (url, error))


async def http_get_text(url: URI, timeout: float = None, encoding: str = 'utf-8') -> Optional[str]:
    """ Get text from URL """
    Log.info(msg='http get text: %s' % url)
    data = await http_get(url=url, allow_redirects=True, timeout=timeout)
    if data is not None:
        return data.decode(encoding=encoding)


async def http_check_m3u8(url: URI, timeout: float = None) -> bool:
    """ Check M3U8 for URL """
    m3u8 = await _http_check(url=url, timeout=timeout)
    if not m3u8:
        Log.warning(msg='this url is not available: %s' % url)
    return m3u8


async def _http_check(url: URI, is_m3u8: bool = None, timeout: float = None) -> bool:
    if is_m3u8 is None and url.lower().find('m3u8') > 0:
        is_m3u8 = True
    status_code, headers = await http_head(url=url, timeout=timeout)
    # check status code
    if status_code == 302 or status_code == 301:
        # redirect
        redirected_url = headers.get('Location')
        return await _http_check(url=redirected_url, is_m3u8=is_m3u8, timeout=timeout)
    elif status_code != 200:
        # HTTP error
        return False
    elif is_m3u8:
        return True
    # check content type
    content_type = headers.get('Content-Type')
    content_type = '' if content_type is None else str(content_type).lower()
    if content_type.find('application/vnd.apple.mpegurl') >= 0:
        return True
    elif content_type.find('application/x-mpegurl') >= 0:
        return True
    Log.warning(msg='this is not a m3u8 url: %s' % url)
    return False


#
#   File
#


def path_parent(path: str) -> str:
    return Path.dir(path=path)


def path_join(parent: str, *children: str) -> str:
    return Path.join(parent, *children)


async def text_file_read(path: str) -> Optional[str]:
    try:
        return await TextFile(path=path).read()
    except Exception as error:
        Log.error(msg='failed to read file: %s, error: %s' % (path, error))


async def text_file_write(text: str, path: str) -> bool:
    try:
        return await TextFile(path=path).write(text=text)
    except Exception as error:
        Log.error(msg='failed to write file: %s (%d bytes), error: %s' % (path, len(text), error))


async def json_file_read(path: str) -> Union[Dict, List, None]:
    text = await text_file_read(path=path)
    if text is None or len(text) < 2:
        return None
    return json_decode(string=text)


async def json_file_write(container: Union[Dict, List], path: str) -> bool:
    text = json_encode(obj=container)
    return await text_file_write(text=text, path=path)



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Any, List, Dict
from typing import Iterable

from ..types import MapInfo
from .stream import LiveStream


class LiveChannel(MapInfo):
    """ TV Channel """

    def __init__(self, info: Dict = None, name: str = None):
        """
        Create Live Channel

        :param info: channel info
        :param name: channel name
        """
        super().__init__(info=info)
        if name is not None:
            self.set(key='name', value=name)

    # Override
    def __str__(self) -> str:
        cname = self.__class__.__name__
        streams = self.get(key='streams', default=[])
        count = len(streams)
        return '<%s name="%s" count=%d />' % (cname, self.name, count)

    # Override
    def __repr__(self) -> str:
        cname = self.__class__.__name__
        streams = self.get(key='streams', default=[])
        count = len(streams)
        return '<%s name="%s" count=%d />' % (cname, self.name, count)

    @property
    def name(self) -> str:
        """ channel name """
        return self.get(key='name', default='')

    @property
    def empty(self) -> bool:
        array = self.get(key='streams', default=None)
        return array is None or len(array) == 0

    @property
    def available(self) -> bool:
        """ source stream(s) available """
        sources = self.streams
        for src in sources:
            if src.available:
                return True
        # channel not available
        return False

    @property
    def streams(self) -> List[LiveStream]:
        """ channel sources """
        array = self.get(key='streams', default=[])
        return LiveStream.convert(array=array)

    @streams.setter
    def streams(self, array: Iterable[LiveStream]):
        value = LiveStream.revert(streams=array)
        self.set(key='streams', value=value)

    def add_stream(self, stream: LiveStream):
        """ add the stream in this channel """
        all_streams = self.streams
        for old in all_streams:
            if old.url == stream.url:
                # same source,
                # update info?
                return False
        all_streams.append(stream)
        self.streams = all_streams
        return True

    def add_streams(self, streams: Iterable[LiveStream]):
        """ add streams in this channel """
        all_streams = self.streams
        count = 0
        for item in streams:
            found = False
            for old in all_streams:
                if old.url == item.url:
                    # same source,
                    # update info?
                    found = True
                    break
            if not found:
                all_streams.append(item)
                count += 1
        if count > 0:
            self.streams = all_streams
            return True

    #
    #   Factories
    #

    @classmethod
    def parse(cls, info: Any):  # -> Optional[LiveChannel]:
        if info is None:
            return None
        elif isinstance(info, LiveChannel):
            return info
        elif isinstance(info, MapInfo):
            info = info.dictionary
        # if 'name' in info:
        #     return cls(info=info)
        return _channel_factory().create_channel(info=info)

    @classmethod
    def convert(cls, array: Iterable[Dict]):  # -> List[LiveChannel]:
        channels: List[LiveChannel] = []
        for item in array:
            info = cls.parse(info=item)
            if info is not None:
                channels.append(info)
        return channels

    @classmethod
    def revert(cls, channels: Iterable) -> List[Dict]:
        array: List[Dict] = []
        for item in channels:
            if isinstance(item, MapInfo):
                array.append(item.dictionary)
            elif isinstance(item, Dict):
                array.append(item)
        return array


def _channel_factory():
    from .factory import LiveFactory
    return LiveFactory()



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Any, List, Dict
from typing import Iterable

from ..types import MapInfo
from .channel import LiveChannel


class LiveGenre(MapInfo):
    """ Channel Group """

    def __init__(self, info: Dict = None, title: str = None):
        """
        Create Channel Group

        :param info:  group info
        :param title: group name
        """
        super().__init__(info=info)
        if title is not None:
            self.set(key='title', value=title)

    # Override
    def __str__(self) -> str:
        cname = self.__class__.__name__
        channels = self.get(key='channels', default=[])
        count = len(channels)
        return '<%s title="%s" count=%d />' % (cname, self.title, count)

    # Override
    def __repr__(self) -> str:
        cname = self.__class__.__name__
        channels = self.get(key='channels', default=[])
        count = len(channels)
        return '<%s title="%s" count=%d />' % (cname, self.title, count)

    @property
    def title(self) -> str:
        """ group name """
        return self.get(key='title', default='')

    @property
    def empty(self) -> bool:
        array = self.get(key='channels', default=None)
        return array is None or len(array) == 0

    @property
    def channels(self) -> List[LiveChannel]:
        """ live channels """
        array = self.get(key='channels', default=[])
        return LiveChannel.convert(array=array)

    @channels.setter
    def channels(self, array: Iterable[LiveChannel]):
        value = LiveChannel.revert(channels=array)
        self.set(key='channels', value=value)

    def add_channel(self, channel: LiveChannel):
        """ add the channel in this genre """
        all_channels = self.channels
        for old in all_channels:
            if old.name == channel.name:
                # same channel, merge streams
                old.add_streams(streams=channel.streams)
                return True
        all_channels.append(channel)
        self.channels = all_channels
        return True

    def add_channels(self, channels: Iterable[LiveChannel]):
        """ all channels in this genre """
        all_channels = self.channels
        count = 0
        for item in channels:
            found = False
            for old in all_channels:
                if old.name == item.name:
                    # same channel, merge streams
                    old.add_streams(streams=item.streams)
                    found = True
                    break
            if not found:
                all_channels.append(item)
                count += 1
        if count > 0:
            self.channels = all_channels
            return True

    #
    #   Factories
    #

    @classmethod
    def parse(cls, info: Any):  # -> Optional[LiveGenre]:
        if info is None:
            return None
        elif isinstance(info, LiveGenre):
            return info
        elif isinstance(info, MapInfo):
            info = info.dictionary
        # if 'title' in info:
        #     return cls(info=info)
        return _genre_factory().create_genre(info=info)

    @classmethod
    def convert(cls, array: Iterable[Dict]):  # -> List[LiveGenre]:
        groups: List[LiveGenre] = []
        for item in array:
            info = cls.parse(info=item)
            if info is not None:
                groups.append(info)
        return groups

    @classmethod
    def revert(cls, genres: Iterable) -> List[Dict]:
        array: List[Dict] = []
        for item in genres:
            if isinstance(item, MapInfo):
                array.append(item.dictionary)
            elif isinstance(item, Dict):
                array.append(item)
        return array


def _genre_factory():
    from .factory import LiveFactory
    return LiveFactory()



# -*- coding: utf-8 -*-
#
#   TV-Box: Live Stream
#
#                                Written in 2024 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Union, Any, List, Dict

from .types import URI, MapInfo
from .utils import md5, utf8_encode, hex_encode
from .utils import path_join
from .utils import json_file_read


def hex_md5(data: Union[bytes, str]) -> str:
    if isinstance(data, str):
        data = utf8_encode(string=data)
    return hex_encode(data=md5(data=data))


class LiveConfig(MapInfo):

    SECTION_NAME = 'tvbox'

    def get_section(self, section: str) -> Dict[str, Any]:
        return self.get(key=section, default={})

    def get_option(self, option: str, section: str = None) -> Optional[Any]:
        if section is None:
            section = self.SECTION_NAME
        sub = self.get_section(section=section)
        return sub.get(option)

    #
    #   Source URLs
    #

    @property
    def sources(self) -> List[URI]:
        array = self.get_option(option='sources')
        return [] if array is None else array

    @property
    def lives(self) -> List[URI]:
        array = self.get_option(option='lives')
        return [] if array is None else array

    @property
    def base_url(self) -> Optional[URI]:
        """ base url """
        return self.get_option(option='base-url')

    @property
    def output_dir(self) -> Optional[str]:
        """ output directory """
        return self.get_option(option='output-dir')

    #
    #   Output file names
    #

    def get_index_file(self) -> str:
        file: str = self.get_option(option='index-file')
        if file is None:
            return 'tvbox.json'
        else:
            return file

    def get_lives_file(self, url: URI) -> str:
        file: str = self.get_option(option='lives-file')
        if file is None or file.find(r'{HASH}') < 0:
            file = 'lives-{HASH}.txt'
        digest = hex_md5(data=url)
        return file.replace(r'{HASH}', digest)

    def get_source_file(self, url: URI) -> str:
        file: str = self.get_option(option='source-file')
        if file is None or file.find(r'{HASH}') < 0:
            file = 'source-{HASH}.txt'
        digest = hex_md5(data=url)
        return file.replace(r'{HASH}', digest)

    #
    #   Output file paths
    #

    def get_output_index_path(self) -> Optional[str]:
        """ output index """
        base = self.output_dir
        if base is not None:
            file = self.get_index_file()
            return path_join(base, file)

    def get_output_lives_path(self, url: URI) -> Optional[str]:
        """ output lives """
        base = self.output_dir
        if base is not None:
            file = self.get_lives_file(url=url)
            return path_join(base, file)

    def get_output_source_path(self, url: URI) -> Optional[str]:
        base = self.output_dir
        if base is not None:
            file = self.get_source_file(url=url)
            return path_join(base, file)

    #
    #   Output URLs
    #

    def get_output_index_url(self) -> Optional[URI]:
        base = self.base_url
        if base is not None:
            file = self.get_index_file()
            return path_join(base, file)

    def get_output_lives_url(self, url: URI) -> Optional[URI]:
        base = self.base_url
        if base is not None:
            file = self.get_lives_file(url=url)
            return path_join(base, file)

    #
    #   Factory
    #

    @classmethod
    async def load(cls, path: str):
        info = await json_file_read(path=path)
        if info is not None:
            return cls(info=info)
